'''
Created on 25.12.2015

@author: michi
'''
import math
import logging

log = logging.getLogger(__name__)

def findRootSecant(func,y,xleft,xright,tol = 1.0e-12, abstol = 1.0e-16, xtol=0, contraction=(0.25,0.5,0.75,0.875,0.9375,0.96875), maxiter = 60):
    '''
    Solve the equation y = func(x) by the secant method.
    
    :param func: The function to evaluate.
    :param y: The ordinate value to search.
    :param xleft: The left point of the search interval
    :param xright: The right point of the search interval
    :param tol: The tolerance for the ordinate value to attain.
    :param abstol: The tolerance near zero for the ordinate value to attain.
    :param xtol: The tolerance for the abscissa value. Needed, if the input function is discrete, i.e. for time function with millisecond precision.
    :param contraction: A list of contraction values to try, if the interval is repeatedly reduced on a single side.
    :param maxiter: The maximal number of iterations to try.
    '''
    
    x1 = xleft
    dy1 = func(x1)-y
    x2 = xright
    dy2 = func(x2)-y
    
    if dy1*dy2 > 0:
        raise ValueError("Initial interval does not have values with opposite signs w.r.t. y.")
    
    if dy1 == 0:
        return x1
    
    if dy2 == 0:
        return x2
    
    if math.fabs(y) < abstol:
        realtol = abstol
    else:
        realtol = math.fabs(y) * tol
    
    repeat = 0
    
    for niter in range(maxiter):
        
        # y2         +
        #           /
        #          /
        # y ------/-------
        #        /
        #       /
        #      /
        # y1  +
        #
        #     x1  xm x2
        #
        #
        # (xm - x1) / (y-y1) = (x2-xm)/(y2-y)
        # (x1 -xm) / dy1 = (x2-xm)/dy2
        # dy2 * (x1-xm) = dy1 * (x2-xm)
        # xm*(dy1-dy2) = dy1 * x2 - dy2 * x1
        # xm = (dy1 * x2 - dy2 * x1) / (dy1-dy2)
    
        xm = (dy1 * x2 - dy2 * x1) / (dy1-dy2)
        
        if math.fabs(x2-x1) <= xtol:
            if log.getEffectiveLevel() <= logging.DEBUG:
                log.debug("x1,xm,x2 [%.16lg,%.16lg,%.16lg]."%(x1,xm,x2))
            return xm
        
        dym = func(xm)-y
        
        if log.getEffectiveLevel() <= logging.DEBUG:
            log.debug("x1,x2 [%.16lg,%.16lg]."%(x1,x2))
            log.debug("findRootSecant: niter,dym,repeat=%d,%lg,%d"%(niter,dym,repeat))
        
        if math.fabs(dym)<realtol:
            return xm
        
        if dy1*dym < 0:
            x2 = xm
            dy2 = dym
            if repeat >= 0:
                repeat += 1
                if repeat > 1:
                    xc1 = x1 + (xm-x1) * contraction[(repeat-2)%len(contraction)]
                    dyc1 = func(xc1) - y
                    if log.getEffectiveLevel() <= logging.DEBUG:
                        log.debug("xc1,dyc1,dym [%.16lg,%.16lg,%.16lg]."%(xc1,dyc1,dym))

                    if dyc1*dym < 0:
                        x1 = xc1
                        dy1 = dyc1
                        repeat = 0
            else:
                repeat = 0
            
        elif dy2*dym < 0:
            x1 = xm
            dy1 = dym
            if repeat <= 0:
                repeat -= 1
                if repeat < -1 and -2-repeat < len(contraction):
                    xc2 = x2 + (xm-x2) * contraction[(-2-repeat)%len(contraction)]
                    dyc2 = func(xc2) - y
                    if log.getEffectiveLevel() <= logging.DEBUG:
                        log.debug("xc2,dyc2,dym [%.16lg,%.16lg,%.16lg]."%(xc2,dyc2,dym))

                    if dyc2*dym < 0:
                        x2 = xc2
                        dy2 = dyc2
                        repeat = 0
            else:
                repeat = 0

        else:
            raise ValueError("Root finding diverged all same signs [%lg,%lg,%lg]."%(dy1,dym,dy2))
    
    raise ValueError("Root finding diverged after %d iterations."%maxiter)

    